#ifndef NNOBJECT_H
#define NNOBJECT_H

/** Nihongo Nano class library
 *
 *      NnObject (IuWFNg)
 *         NnSortable (r\IuWFNg)
 *            NnString ()
 *         NnVector  (ϒz)
 *         NnHash (L[̃nbV)
 *         NnBullet  (L[ƒl̃yA)
 *         Writer    (ďo)
 *            FileWriter    (t@Co:NYADOS)
 *            StreamWriter  (FILE*o)
 *               FileWriter (t@Co:NYACUS)
 *               PipeWriter (pCvo)
 *         Reader    (ē)
 *            StreamReader  (FILE*)
 *               FileReader (t@C)
 *               PipeReader (pCv)
 *         NnEnum    (C^[^[)
 *              NnHash::Each (nbV̊evf  NnBullet )
 *              NnDir           (fBNg      NnString )
 *
 * Edelete 錠/`uLvƌĂԁB
 *
 * ERNVnIuWFNg͔i[IuWFNgɑ΂
 *   LLBāAauto/staticñIuWFNg
 *   RNVnIuWFNg֓o^łȂB
 *
 * ENnObject ́AfXgN^zĂA
 *   RNVnIuWFNgɊi[IuWFNg
 *   NnObject hȂĂ͂ȂB
 *
 * EɓnIuWFNgɊւāALړꍇ
 *   |C^nB
 *      ()
 *          NnVector vec;
 *          vec.append( new NnString("hogehoge") );
 */

/* ̐擪oCgł΁A0 Ԃ */
extern char dbcs_table[ 256 + 128 ];
void init_dbcs_table();

#undef  isKanji
#define isKanji(n) dbcs_table[(unsigned)(128+(n))]

#undef  isSpace
#define isSpace(x) isspace((unsigned)((x)&255))
#undef  isDigit
#define isDigit(x) isdigit((unsigned)((x)&255))
#undef  isAlpha
#define isAlpha(x) isalpha((unsigned)((x)&255))
    
/* ÓIz̗vfԂ */
#undef  numof
#define numof(X) (sizeof(X)/sizeof((X)[0]))

/* ŏ̒̕lO\萔Ƃ */
enum{ EXCEPTIONS = (int)(~0u-(~0u >> 1)) };

class NnSortable;

/** NnObject NX
 *   fXgN^zNXB
 *   p΁ANnVector , NnHash ɓo^邱ƂłB
 */
class NnObject {
public:
    virtual ~NnObject();
    virtual NnObject   *clone() const;
    virtual NnSortable *sortable();
    virtual const char *repr() const { return "<NnObject>"; }
};

/** NnSortable NX
 *    r\ȃIuWFNg͂ꂩpƁA
 *    NnVector ŁAsort,uniq \bhgƂłB
 */
class NnSortable : public NnObject {
public:
    /** x ƔrāA召֌W EE̐lŕԂB*/
    virtual int compare( const NnSortable &x) const = 0;
    virtual NnSortable *sortable();
};

/** NnEnum NX
 *    vf񋓂C^[tFCX
 */
class NnEnum : public NnObject {
public:
    /** ̗vfֈړ */
    virtual void      operator ++ ()=0;
    void next(){ ++*this; }

    /** ݍĂvfւ̃|C^Ԃ.
     *  SĂ̗vf𑖍IƁANULL Ԃ.
     */
    virtual NnObject *operator *  ()=0;

    /** ܂SĂ̗vf𑖍ĂȂ ^Ԃ */
    virtual int more();
};

/* ̃wb_Ő錾Ă郁\bh̎̂
 * NnString.cpp Œ`ĂB
 */

template <class T> T max1(T a,T b){ return a<b ? b : a ; }
template <class T> T min1(T a,T b){ return a<b ? a : b ; }

const char *getEnv( const char *var , const char *none=NULL );

class NnVector;
class NnExecutable : public NnObject{
public:
    virtual int operator()( const NnVector &args )=0;
};

#endif

