#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <process.h>
#include <io.h>
#include "shell.h"
#include "nnstring.h"
#include "nnhash.h"
#include "getline.h"
#include "reader.h"

static void glob( NnString &line )
{
    if( properties.get("glob") ){
        NnString result;
        glob_all( line.chars() , result );
        line = result;
    }
}

/* tNH[gœǂ݂߂镶ʂ߂ */
static int getQuoteMax()
{
    NnString *max=(NnString*)properties.get("backquote");
    return max == NULL ? 0 : 8192 ;
}

#ifdef THREAD

struct BufferingInfo {
    NnString buffer;
    int fd;
    int max;
    BufferingInfo(int f,int m) : fd(f),max(m){}
};

static void buffering_thread( void *b_ )
{
    BufferingInfo *b= static_cast<BufferingInfo*>( b_ );
    char c;

    while( ::read(b->fd,&c,1) > 0 ){
        if( b->buffer.length()+1 < b->max ){
            b->buffer << c;
        }
    }
    ::close( b->fd );
}

#endif

/* R}h̕Wo͂̓e𕶎Ɏ荞. 
 *   cmdline - sR}h
 *   dst - Wo͂̓e
 *   max - p̏. 0 ̎͏
 *   shrink - ȏ̋󔒕ɂ鎞 true Ƃ
 * return
 *   0  : 
 *   -1 : I[o[t[(max ܂ł dst Ɏ荞ł)
 *   -2 : e|t@C쐬łȂ
 */
int eval_cmdline( const char *cmdline, NnString &dst, int max , bool shrink)
{
#ifdef THREAD
    extern int mkpipeline( int pipefd[] );
    int pipefd[2];

    if( mkpipeline( pipefd ) != 0 )
        return -1;

    /* o͂󂯎~߂Xbhsđ点Ă */
    BufferingInfo buffer(pipefd[0],max);
#ifdef __EMX__
    if( _beginthread( buffering_thread , NULL , 655350u , &buffer ) == -1 )
        return -1;
#else
    if( _beginthread( buffering_thread , 0 , &buffer ) == (uintptr_t)-1L)
        return -1;
#endif

    int savefd = dup(1);
    dup2( pipefd[1] , 1 );
    ::close(pipefd[1]);
#else
    char *tmppath=NULL;
    if( (tmppath=_tempnam("\\","NYA"))==NULL ){
        return -2;
    }
    FILE *fp=fopen(tmppath,"w+");
    if( fp==NULL ){
        free(tmppath);
        return -2;
    }
    int savefd = dup(1);
    dup2( fileno(fp) , 1 );
#endif

    OneLineShell shell( cmdline );
    shell.mainloop();

    dup2( savefd , 1 );
    int lastchar=' ';
#ifdef THREAD
    for( const char *p=buffer.buffer.chars() ; *p != '\0' ; ++p ){
        if( shrink && isSpace(*p) ){
            if( ! isSpace(lastchar) )
                dst += ' ';
        }else{
            dst += (char)*p;
        }
        lastchar = *p;
    }
#else
    rewind( fp );
    int ch;
    int rc=0;
    while( !feof(fp) && (ch=getc(fp)) != EOF ){
        if( shrink && isSpace(ch) ){
            if( ! isSpace(lastchar) )
                dst += ' ';
        }else{
            dst += (char)ch;
        }
        lastchar = ch;
        if( max && dst.length() >= max ){
            rc = -1;
            break;
        }
    }
    fclose(fp);
    if( tmppath != NULL ){
        remove(tmppath);
        free(tmppath);
    }
#endif
    dst.trim();
    return rc;
}

/* tNH[g
 *      sp  (`̎wĂ邱)
 *      dst 敶
 *      max 
 *      quote pɈ͂܂ĂȂ 0 ɃZbgB
 *           (0 ̎A&  > ƂNH[g)
 * return
 *      0  
 *      -1 I[o[t[
 *      -2 e|t@C쐬łȂB
 */
static int doQuote( const char *&sp , NnString &dst , int max , int quote )
{
    NnString q;
    while( *sp != '\0' ){
        if( *sp =='`' ){
            /* A `` ́A ` ֕ϊB */
            if( *(sp+1) != '`' )
                break;
	    ++sp;
        }
	if( isKanji(*sp) )
	    q += *sp++;
        q += *sp++;
    }
    if( q.length() <= 0 ){
        dst += '`';
        return 0;
    }

    NnString buffer;
    int rc=eval_cmdline( q.chars() , buffer , max , !quote );
    if( rc < 0 ){
        return rc;
    }
    if( quote ){
        for(const char *p = buffer.chars() ; *p != '\0' ; ++p ){
            if( *p == '"' ){
                dst << "\"\"";
            }else{
                dst << *p;
            }
        }
    }else{
        for(const char *p = buffer.chars() ; *p != '\0' ; ++p ){
            if( *p == '&' || *p == '|' || *p == '<' || *p=='>' ){
                dst << '"' << *p << '"';
            }else if( *p == '"' ){
                dst << "\"\"\"\"";
            }else{
                dst << *p;
            }
        }
    }
    return rc;
}


/* _CNgǂݎAʂ FileWriter IuWFNgŕԂ.
 * I[vłȂꍇ̓G[bZ[Wɏo.
 *	sp ŏ'>'̒
 *         śA
 * return
 *      o͐
 */
static FileWriter *readWriteRedirect( const char *&sp )
{

    // append ?
    const char *mode = "w";
    if( *sp == '>' ){
	++sp;
	mode = "a";
    }
    NnString fn;
    NyadosShell::readNextWord(sp,fn);
    if( fn.empty() ){
        conErr << "syntax error near unexpected token `newline'\n";
	return NULL;
    }

    FileWriter *fw=new FileWriter(fn.chars(),mode);

    if( fw != NULL && fw->ok() ){
	return fw;
    }else{
	perror( fn.chars() );
	delete fw;
	return NULL;
    }
}

/* sp tH_[񂩂n܂Ă^Ԃ
 *	sp c 擪ʒu
 *	quote c dpȂ^ɂ
 *	spc c O̕󔒂Ȃ^ɂĂ
 * return
 *	doFolder ŕϊׂł΁A̒
 */
int isFolder( const char *sp , int quote , int spc )
{
    if( ! spc  ||  quote )
	return 0;

    if( *sp=='~' && 
        (getEnv("HOME")!=NULL || getEnv("USERPROFILE") != NULL ) && 
        properties.get("tilde")!=NULL)
    {
	int n=1;
	while( isalnum(sp[n] & 255) || sp[n]=='_' )
	    ++n;
	return n;
    }

    if( sp[0]=='.' && sp[1]=='.' && sp[2]=='.' && properties.get("dots")!=NULL ){
	int n=3;
	while( sp[n] == '.' )
	    ++n;
	return n;
    }
    return 0;
}
/* sp Ŏn܂tH_[{̃tH_[֕ϊ
 *	sp c 擪ʒu
 *	len c 
 *	dst c {̃tH_[Ƃ
 *	quote c dpȂ^ɂ
 */	
static void doFolder( const char *&sp , int len , NnString &dst , int & )
{
    NnString name(sp,len),value;
    NnDir::filter( name.chars() , value );
    sp += len;
    if( value.findOf(" \t\v\r\n") >= 0 ){
	dst << '"' << value << '"';
    }else{
	dst << value;
    }
}

/* ϐȂǂWJ(R}h)
 *	src - 
 *	dst - ʂ
 * return
 *      0 -  , -1 - s
 */
int NyadosShell::explode4internal( const NnString &src , NnString &dst )
{
    /* pCv܂̏Ɏs */
    NnString firstcmd; /* ŏ̃pCv܂ł̃R}h */
    NnString restcmd;  /* ڍs̃R}h */
    src.splitTo( firstcmd , restcmd , "|" , "\"`" );
    if( ! restcmd.empty() ){
	if( restcmd.at(0) == '&' ){ /* Wo+G[o */
	    NnString pipecmds( restcmd.chars()+1 );

            PipeWriter *pw=new PipeWriter(pipecmds);
            if( pw == NULL || ! pw->ok() ){
                /* PipeWriter ŃG[o͂Ă̂ŁA
                 * bZ[W͏oȂĂ悢B
                 * perror( pipecmds.chars() );
                 */
                delete pw;
                return -1;
	    }
            conOut_ = pw ;
            conErr_ = new WriterClone(pw);
	}else{ /* Wo͂̂ */
            NnString pipecmds( restcmd.chars() );

            PipeWriter *pw=new PipeWriter(pipecmds);
            if( pw == NULL || ! pw->ok() ){
                /* PipeWriter ŃG[o͂Ă̂ŁA
                 * bZ[W͏oȂĂ悢B
                 * perror( pipecmds.chars() );
                 */
                delete pw;
                return -1;
	    }
            conOut_ = pw;
	}
    }
    int prevchar=' ';
    int quote=0;
    int len;
    dst.erase();
    int backquotemax=getQuoteMax();

    const char *sp=firstcmd.chars();
    while( *sp != '\0' ){

	if( *sp == '"' )
	    quote = !quote;
	
	int prevchar1=(*sp & 255);
	if( (len=isFolder(sp,quote,isSpace(prevchar))) != 0 ){
	    doFolder( sp , len , dst , quote );
        }else if( *sp == '<' && !quote ){
            ++sp;
            NnString fn;
            NyadosShell::readNextWord(sp,fn);
            if( fn.empty() ){
                conErr << "syntax error near unexpected token `newline'\n";
                return -1;
            }
            FileReader *fr=new FileReader(fn.chars());
            if( fr != NULL && !fr->eof() ){
                conIn_ = fr;
            }else{
                conErr << "can't redirect stdin.\n";
                return -1;
            }
        }else if( *sp == '>' && !quote ){
	    ++sp;
	    FileWriter *fw=readWriteRedirect( sp );
	    if( fw != NULL ){
                conOut_ = fw;
	    }else{
		conErr << "can't redirect stdout.\n";
		return -1;
	    }
	}else if( *sp == '1' && *(sp+1) == '>' && !quote ){
	    if( ! restcmd.empty() ){
		/* łɃ_CNgĂAG[ */
		conErr << "ambigous redirect.\n";
		return -1;
	    }
	    sp += 2;
	    if( *sp == '&'  &&  *(sp+1) == '2' ){
		sp += 2;
                conOut_ = new WriterClone(conErr_);
	    }else if( *sp == '&' && *(sp+1) == '-' ){
		sp += 2;
                conOut_ = new NullWriter();
	    }else{
		FileWriter *fw=readWriteRedirect( sp );
		if( fw != NULL ){
                    conOut_ = fw;
		}else{
		    conErr << "can't redirect stdout.\n";
		    return -1;
		}
	    }
	}else if( *sp == '2' && *(sp+1) == '>' && !quote ){
	    sp += 2;
	    if( *sp == '&' && *(sp+1) == '1' ){
		sp += 2;
                conErr_ = new WriterClone(conOut_);
	    }else if( *sp == '&' && *(sp+1) == '-' ){
		sp += 2;
                conErr_ = new NullWriter();
	    }else{
		FileWriter *fw=readWriteRedirect( sp );
		if( fw != NULL ){
                    conErr_ = fw;
		}else{
		    conErr << "can't redirect stderr.\n";
		    return -1;
		}
	    }
        }else if( *sp == '`' && backquotemax > 0 ){
	    ++sp;
            switch( doQuote( sp , dst , backquotemax , quote ) ){
            case -1:
                conErr << "Over flow commandline.\n";
                return -1;
            case -2:
                conErr << "Can not make temporary file.\n";
                return -1;
            default:
                break;
            }
	    ++sp;
	}else{
	    if( isKanji(*sp) )
		dst += *sp++;
	    dst += *sp++;
	}
	prevchar = prevchar1;
    }
    glob( dst );
    return 0;
}

/* qAhLgs
 *    sp - u<<ENDv̍ŏ < ̈ʒuB
 *    dst - (u< ꎞt@Cv)
 *    prefix - u<voru v
 */
void NyadosShell::doHereDocument( const char *&sp , NnString &dst , char prefix )
{
    int quote_mode = 0 , quote = 0 ;
    sp += 2;

    /* I[}[Nǂݎ */
    NnString endWord;
    while( *sp != '\0' && (!isspace(*sp & 255) || quote ) ){
	if( *sp == '"' ){
	    quote ^= 1;
	    quote_mode = 1;
	}else{
	    endWord << *sp ;
	}
	++sp;
    }
    /* hLgꎞt@Cɏo */
    this->heredocfn=NnDir::tempfn();
    FILE *fp=fopen( heredocfn.chars() , "w" );
    if( fp != NULL ){
	VariableFilter  variable_filter(*this);
	NnString  line;

	NnString *prompt=new NnString(endWord);
	*prompt << '>';
	this->nesting.append( prompt );

	while( this->readline(line) >= 0 && !line.startsWith(endWord) ){
	    /* <<"EOF" ł͂ȂA<<EOF `̏ꍇ́A
	     * %...% `̒PuKv
	     */
	    if( ! quote_mode )
		variable_filter( line );

	    fputs( line.chars() , fp );
	    if( ! line.endsWith("\n") )
		putc('\n',fp);
	    line.erase();
	}
	delete this->nesting.pop();
	fclose( fp );
	dst << prefix << heredocfn ;
    }
}


/* OR}hpvvZX.
 * E%1,%2 ϊ
 * EvO /  ֕ϊB
 *      src - 
 *      dst - Hʕ̓
 * return 0 -  , -1 s(I[o[t[Ȃ)
 */
int NyadosShell::explode4external( const NnString &src , NnString &dst )
{
    /* vOtB^[ɂF`_ϊȂ */
    NnString progname,args;
    src.splitTo( progname,args );
    NnDir::filter( progname.chars() , dst );

    /* vOɋ󔒂܂܂ĂA"" ň͂ݒ */
    if( dst.findOf(" ") >= 0 ){
        dst.dequote();
        dst.unshift('"');
        dst << '"';
    }

    if( args.empty() )
	return 0;

    dst << ' ';
    int backquotemax=getQuoteMax();
    int quote=0;
    int len;

    // ̃Rs[.
    int spc=1;
    for( const char *sp=args.chars() ; *sp != '\0' ; ++sp ){
        if( *sp == '"' )
            quote ^= 1;

	if( (len=isFolder(sp,quote,spc)) != 0 ){
	    doFolder( sp , len , dst , quote );
	    --sp;
        }else if( *sp == '`'  && backquotemax > 0 ){
	    ++sp;
            switch( doQuote( sp , dst , backquotemax , quote ) ) {
            case -1:
                conErr << "Over flow commandline.\n";
                return -1;
            case -2:
                conErr << "Can not make temporary file.\n";
                return -1;
            default:
                break;
            }
	}else if( *sp == '<' && *(sp+1) == '<'  &&  quote == 0 ){
	    /* qAhLg */
	    doHereDocument( sp , dst , '<' );
	}else if( *sp == '<' && *(sp+1) == '=' && quote == 0 ){
	    /* CCt@CeLXg */
	    doHereDocument( sp , dst , ' ' );
	}else if( *sp == '|' && *(sp+1) == '&' ){
	    dst << "2>&1 |";
        }else{
	    if( isKanji(*sp) )
		dst += *sp++;
            dst += *sp;
        }
        spc = isSpace( *sp );
    }
    glob( dst );
    return 0;
}
